#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#include "ipc_msg.h"

struct mymsgbuf ipc_qbuf;

int ipc_init(key_t key)
{
	int msgqueue_id;

        /* Open the queue - create if necessary */
        if((msgqueue_id = msgget(key, IPC_CREAT|0666)) == -1) 
	{
                perror("msgget");
		exit(1);
        }
	return msgqueue_id;
}

void ipc_send_message(int qid, long type, char *text)
{
        /* Send a message to the queue */

        ipc_qbuf.mtype = type;
        strcpy(ipc_qbuf.mtext, text);

        if((msgsnd(qid, (struct msgbuf *)&ipc_qbuf,
                strlen(ipc_qbuf.mtext)+1, 0)) ==-1)
        {
                perror("msgsnd");
                exit(1);
        }
}

void ipc_read_message(int qid, long type)
{
        /* Read a message from the queue */
        ipc_qbuf.mtype = type;
        msgrcv(qid, (struct msgbuf *)&ipc_qbuf, MAX_SEND_SIZE, type, 0);
}

void ipc_read_message_nowait(int qid, long type)
{
        /* Read a message from the queue */
        ipc_qbuf.mtype = type;
        msgrcv(qid, (struct msgbuf *)&ipc_qbuf, MAX_SEND_SIZE, type, IPC_NOWAIT); 
}        


void ipc_remove_queue(int qid)
{
        /* Remove the queue */
        msgctl(qid, IPC_RMID, 0);
}

void ipc_change_queue_mode(int qid, char *mode)
{
        struct msqid_ds myqueue_ds;

        /* Get current info */
        msgctl(qid, IPC_STAT, &myqueue_ds);

        /* Convert and load the mode */
        sscanf(mode, "%ho", &myqueue_ds.msg_perm.mode);

        /* Update the mode */
        msgctl(qid, IPC_SET, &myqueue_ds);
}

int ipc_peek_message(int qid, long type)
{
        int     result;

        if((result = msgrcv( qid, NULL, 0, type,  IPC_NOWAIT)) == -1)
        	return(TRUE);
        
        return(FALSE);
}
