/*
 *   ntalk v1.0.0 by Sami Kystil, March 1999
 *
 *   A multi-user chat program.
 */

#include <time.h>
#include <curses.h>
#include <term.h>
#include <signal.h>

#include "ipc_msg.h"
#include "ipc_shm.h"
#include "vt100.h"
#include "ntalk.h"
#include "ttt.h"

#define EXEC_TMP "/var/tmp/nt_tmp.%d"

#define MSG_VERSION	"*** NTalk v"T_VERSION" by Sami Kystil, March 1999\r\n"
#define MSG_HELP	"*** Commands:\r\n"\
			"*** /bye, /quit, /exit, /leave	- exit the "ROOMNAME"\r\n"\
			"*** /w, /who, /names		- display list of users in the "ROOMNAME"\r\n"\
			"*** /invite <name> <reason>	- invite a user to the "ROOMNAME"\r\n"\
			"***				  (both users must have messages enabled)\r\n"\
			"*** /msg <name> <message>	- send a private message to a user\r\n"\
			"*** /exec <command>, /e <cmd>	- spawns a shell to exec the specified command\r\n"\
			"***				  (messages received while the process is\r\n"\
			"***				  running are buffered and displayed after\r\n"\
			"***				  the task is complete)\r\n"\
			"*** /help			- this message\r\n"

#define SCREEN_LINES (lines)


int	qid, smid;
char	u_lname[16];
short	u_tid;
char	u_idstring[32];
char	u_input[MAX_INPUT_LENGTH];
int	u_input_pos=0;

user_t	*r_user;
user_t  me;
int	me_index;
int	t_mode;
int	last_msg_t;
char	*last_msg = ipc_qbuf.mtext;


void make_sure_the_fools_know_im_still_around(void)
{
	r_user[me_index].tid = me.tid;
}

int exec_cmd(char *cmd)
{
	char tmp[256], buffer[256];

	FILE *f;
	int i,j,k;

	memset(tmp, 0, 256);
	sprintf(tmp, EXEC_TMP, me.tid);
	sprintf(buffer, "%s > %s", cmd, tmp);

	k = system(buffer);

	cursor_set(1, SCREEN_LINES);
	printf(VT_ERASE);

	i=0;
	f = fopen(tmp, "rb");
	while (!feof(f))
	{
		j = fgetc(f);
		printf("%c", j);
		if (j=='\n') {printf("\r"); i++;}
		if (i==SCREEN_LINES-2)
		{
			printf("*** more...\r");
			fflush(stdout);
			while(i!=' '&&i!=13)
			{
				i=getch();
				make_sure_the_fools_know_im_still_around();
			}
			printf(VT_ERASE);
			i=0;
		}
	}
	fclose(f);
	unlink(tmp);
	return k;
}

void t_leave(void)
{
	char buffer[256]; 
	sprintf(buffer, "%s/%d", me.name, me.tid);
	msg_send(M_LEAVE, buffer);

	r_user[me_index].tid = 0;
	r_user[me_index].name[0] = '\0';

	if (count_users()==0)
	{
		ipc_remove_queue(qid);
		ipc_remove_shm(smid);
	}
	sprintf(buffer, "*** %s/%d has left the %s.", me.name, me.tid, ROOMNAME);
	t_printmsg(buffer); 
}

void str_lcase(char *str)
{
	int i;
	for (i=0; i<strlen(str); i++)
		if (str[i]>='A' && str[i]<='Z') str[i]+=32;
}

void delay(long ticks)
{
	long start = clock();
	while(clock() < start+ticks);
}

int count_users()
{
	int i, j=0;
	for (i=0; i<MAX_USERS; i++)
		if (r_user[i].tid) j++;

	return j;
}

int user_index(char *name)
{
	int i;
	for (i=0; i<MAX_USERS; i++)
		if (strcmp(r_user[i].name, name)==0 && r_user[i].tid) return i;

	return -1;
}

void remove_dead_users()
{
	int i;
        user_t *user_temp = (user_t*)malloc(MAX_USERS*sizeof(user_t));

        for (i=0; i<MAX_USERS; i++)
                if (r_user[i].tid!=0)
			r_user[i].tid = 0;

	delay(CLOCKS_PER_SEC/2);
}

void t_print_users()
{
	char buffer[4096];
	int i;

	remove_dead_users();
	r_user[me_index].tid = me.tid;

	sprintf(buffer, "*** %d user(s): ", count_users());

	for (i=0; i<MAX_USERS; i++)
	{
		if (r_user[i].tid)
		{
			strcat(buffer, r_user[i].name);
			strcat(buffer, " ");
		}
	}

	t_printmsg(buffer);
}

void t_print_input()
{
	cursor_set(1, SCREEN_LINES);
	if (u_input_pos<80)
		printf(u_input);
	else
		printf(&u_input[u_input_pos-80]);
}


void t_printmsg(char *text)
{
	int i, j;
/*
	if (t_mode == MODE_NORMAL)
	{
*/
		cursor_set(1, SCREEN_LINES);
		printf(VT_ERASE);
		for (i=0; i<strlen(text); i++, j++)
		{
			printf("%c", text[i]);
			if (j==80)
			{
				j=0;
				printf("\r\n%s", VT_ERASE);
			}
		}
		printf("\r\n");
		t_print_input();
/*
	} else if (t_mode == MODE_TTT)
	{
		cursor_set(1, SCREEN_LINES);
		printf(VT_ERASE);
		for (i=0; i<80; i++)
			printf("%c", text[i]);
		t_print_input();
	}
*/
	refresh();
}

int t_input()
{
	char key = getch();
	char buffer[300];

	if (key && key != ERR)
	{
		if (key == 13 && u_input_pos>0)
		{
			key = 0;

			if (u_input[0] == CMDCHAR)
			{
				key = t_parsecmd(&u_input[1]);
				if (key) return key;
				key = 0;
			} else {
				sprintf(buffer, "<%s> %s", me.name, u_input);
				msg_send(M_MSG, buffer);

        	                sprintf(buffer, "%s<%s>%s %s", VT_BOLD, me.name, VT_NORMAL, u_input);
        	                msg_sendpriv(me_index, M_MSG, buffer);
			}
			u_input[0] = '\0';
			u_input_pos = 0;
			memset(u_input, 0, MAX_INPUT_LENGTH);
			cursor_set(1, SCREEN_LINES);
			printf(VT_ERASE);
		}
		if (key == 13) key = 0; 
		if (key == 8 || key == 127)
		{
			key = 0;
			if (u_input_pos > 0) u_input[--u_input_pos] = '\0';
			u_input[u_input_pos] = '\0';
			cursor_set(u_input_pos+1, SCREEN_LINES);
			printf(VT_ERASE);
		}

		if (key) u_input[u_input_pos++] = key;
		t_print_input();
		refresh();
	}
	return 0;
}

void msg_send(long type, char *data)
{
	int i;
	for (i=0; i<MAX_USERS; i++)
		if (i != me_index && r_user[i].tid)
	                ipc_send_message(qid, r_user[i].tid*USER_SCALAR+type, data);
}

void msg_broadcast(long type, char *data)
{
        int i;
        for (i=0; i<MAX_USERS; i++)
		if (r_user[i].tid)
			ipc_send_message(qid, r_user[i].tid*USER_SCALAR+type, data);
}


void msg_sendpriv(int user, long type, char *data)
{        
	if (r_user[user].tid)
		ipc_send_message(qid, r_user[user].tid*USER_SCALAR+type, data);
}



int t_parsecmd(char *cmd)
{
	char command[256];
	char buffer[256];
	char parm1[256];

	int i, j=0, k;

	for (i=0; i<strlen(cmd); i++)
	{
		if (cmd[i]==' ') break;
		command[i] = cmd[i];
	}
	command[i] = '\0';

	if (!strlen(command)) return 0;

	i++;
        for (;i<strlen(cmd); i++)
        {
                if (cmd[i]==' ') break;
                parm1[j++] = cmd[i];   
        }
        parm1[j] = '\0';

	str_lcase(command);

	if (strcmp(command, "quit")==0 || 
	    strcmp(command, "bye")==0 || 
	    strcmp(command, "leave")==0 || 
	    strcmp(command, "exit")==0)
		return 1;

	if (strcmp(command, "who")==0 || 
	    strcmp(command, "names")==0 ||
	    strcmp(command, "w")==0)
	{
		t_print_users();
		return 0;
	}

	if (strcmp(command, "help")==0)
	{
		cursor_set(1, SCREEN_LINES);
		printf(VT_ERASE);
		printf(MSG_VERSION);
		printf(MSG_HELP);	
		return 0;
	}

	if (strcmp(command, "exec")==0)
	{
		if (strlen(cmd)>=5)
		{
			sprintf(buffer, "*** executing '%s'", &cmd[5]);
			t_printmsg(buffer);

			k = exec_cmd(&cmd[5]);
			sprintf(buffer, "*** returned %d (0x%X)", k, k);
			t_printmsg(buffer);
		} else
			t_printmsg("*** usage: exec <command>");
		return 0;
	}
	if (strcmp(command, "e")==0)
	{
		if (strlen(cmd)>=2)
		{
			sprintf(buffer, "*** executing '%s'", &cmd[2]);
			t_printmsg(buffer);

			k = exec_cmd(&cmd[2]);
			sprintf(buffer, "*** returned %d (0x%X)", k, k);
			t_printmsg(buffer);
		} else
			t_printmsg("*** usage: e <command>");
		return 0;
	}

	if (strcmp(command, "msg")==0 && i<strlen(cmd))
	{
		k = user_index(parm1);
		if (k == -1)
			t_printmsg("*** Invalid user");
		else
		{
			sprintf(buffer, "*%s* %s", me.name, &cmd[i+1]);
			msg_sendpriv(k, M_MSG, buffer);

                        sprintf(buffer, "-> *%s* %s", parm1, &cmd[i+1]);
			t_printmsg(buffer);
		}
		return 0;
	}

        if (strcmp(command, "invite")==0)
	{
		if (strlen(parm1))
		{
			sprintf(buffer, "*** Inviting %s to the %s...", parm1, ROOMNAME);
			t_printmsg(buffer);

			sprintf(buffer, "echo \"\r\n"
					"*** %s has invited you to join NTalk.\r\n"
					"*** %s\r\n"
					"*** Join by typing 'nt' in the command prompt or\r\n"
					"*** disable this message with 'mesg n'.\r\n"
					"\" | write %s > /dev/null",
					me.name, &cmd[i+1], parm1);
			if (system(buffer) != 0)
				t_printmsg("*** write returned bad status.");
		} else
			t_printmsg("*** usage: invite <name> <reason>");
		return 0;
	}

/*
	if (strcmp(command, "tttjoin")==0)
	{
		
	}

	if (strcmp(command, "ttt")==0)
	{
		if (!strlen(parm1))
		{
			sprintf(buffer, "*** %s has started a ttt server, join by typing /ttt %s", me.name, me.name);
			msg_send(M_MSG, buffer);
			r_user[me_index].mode = UM_TTT_SERVER;
			
			t_printmsg("*** Waiting for a client, press 'q' to abort.");
			last_msg_t = 0;
			while (last_msg_t != M_TTTJOIN)
			{
				k = getch();
				if (k=='q' || k=='Q')
				{
					t_printmsg("*** ttt: aborted");
					break;
				}
				msg_handle();
			}

			if (last_msg_t == M_TTTJOIN)
			{
				sprintf(buffer, "*** %s has joined.", last_msg);
				t_printmsg(buffer);
				ttt_play(1, me_index, user_index(last_msg));
			}
		} else
		{
			k = user_index(parm1);

			if (k != -1)
			{
				if (r_user[k].mode == UM_TTT_SERVER)
				{
					sprintf(buffer, "*** Joined ttt game hosted by %s.", r_user[k].name);
					t_printmsg(buffer);
					msg_sendpriv(k, M_TTTJOIN, me.name);
	                                ttt_play(0, me_index, k);
				} else
				{
					sprintf(buffer, "*** %s is not running a ttt server.", r_user[k].name);
					t_printmsg(buffer);
				}
			} else
				t_printmsg("*** Invalid user");
		}
		return 0;
	}
*/
/*
	if (strcmp(command, "ttt")==0)
	{
		if (strlen(parm1))
		{
			for (i=0; i<MAX_USERS+1; i++)
			{
				if (i==MAX_USERS)
				{
					t_printmsg("*** Invalid user");
					return;
				}

				if (strcmp(r_user[i].name, parm1)==0) break;

			}
			sprintf(buffer, "*** Inviting %s (%d) for a round of TTT...", parm1, i);
			t_printmsg(buffer);			
		} else
			t_printmsg("*** usage: ttt user");
	}
*/
	sprintf(buffer, "*** Unknown command: %s", command);
	t_printmsg(buffer);
	return 0;
}


int msg_handle()
{
	int	i;
	char	buffer[256];

	make_sure_the_fools_know_im_still_around();

        ipc_qbuf.mtext[0] = '\0';
	ipc_read_message_nowait(qid, me.tid*USER_SCALAR + M_MSG);
	if (strlen(ipc_qbuf.mtext))
	{
		last_msg_t = M_MSG;
		t_printmsg(ipc_qbuf.mtext);
		return 0;
	}

        ipc_qbuf.mtext[0] = '\0';
        ipc_read_message_nowait(qid, me.tid*USER_SCALAR + M_TTTJOIN);
        if (strlen(ipc_qbuf.mtext))
        {
                last_msg_t = M_TTTJOIN;
		return 0;	
	}

        ipc_qbuf.mtext[0] = '\0';
        ipc_read_message_nowait(qid, me.tid*USER_SCALAR + M_TTTMOVE);
        if (strlen(ipc_qbuf.mtext))
        {
                last_msg_t = M_TTTMOVE;
                return 0;
        }


        ipc_qbuf.mtext[0] = '\0';
        ipc_read_message_nowait(qid, me.tid*USER_SCALAR + M_TTTMOVE);  
        if (strlen(ipc_qbuf.mtext))
        {
                last_msg_t = M_MSG;  
        	return 0;
	}


        ipc_qbuf.mtext[0] = '\0';
        ipc_read_message_nowait(qid, me.tid*USER_SCALAR + M_ENTER);
        if (strlen(ipc_qbuf.mtext)) 
	{
		last_msg_t = M_ENTER;
		sprintf(buffer, "*** %s has entered the %s.", ipc_qbuf.mtext, ROOMNAME);
		t_printmsg(buffer);
		return 0;
	}

        ipc_qbuf.mtext[0] = '\0';
        ipc_read_message_nowait(qid, me.tid*USER_SCALAR + M_LEAVE);
        if (strlen(ipc_qbuf.mtext)) 
	{
		last_msg_t = M_LEAVE;
		sprintf(buffer, "*** %s has left the %s.", ipc_qbuf.mtext, ROOMNAME);
		t_printmsg(buffer);
		return 0;
	}

	return 0;
}

void sig_int()
{
	t_leave();        
	nocbreak(); echo();
	endwin();

        r_user[me_index].tid = 0;
        r_user[me_index].name[0] = '\0';

	exit(1);  
}
