#include "ttt.h"

char grid[GRIDXSIZE][GRIDYSIZE];
char *keyset = "adws";
int mark_x=GRIDXSIZE/2, mark_y=GRIDYSIZE/2;

void update_userdata(char *user, int win, int lose, int quit)
{
	FILE *handle = fopen(USERFILE, "r+");
		
	int i, found=0;
	long offset=0;

	user_data u;
	memset(&u, 0, sizeof(user_data));

        if (handle == NULL)
        {
                t_printmsg("Unable to update data file.");
                return;
        }

	while (!feof(handle))
	{
		offset = ftell(handle);
		fread(&u, sizeof(user_data), 1, handle);
		if (strcmp(u.name, user)==0)
		{
			found=1;
			break;
		}
	}
	if (!found)
	{
		fseek(handle, 0, SEEK_END);
		offset = ftell(handle);
		memset(&u, 0, sizeof(user_data));
		u.win = 0;
		u.lose = 0;
		u.quit = 0;
	}

	fseek(handle, offset, SEEK_SET);
	strcpy(u.name, user);
	u.win += win;
	u.lose += lose;
	u.quit += quit;
	fwrite(&u, sizeof(user_data), 1, handle);	
	
	fclose(handle);
}

void view_top()
{
        FILE *handle = fopen(USERFILE, "r");

        int i=0, j, k, total;
        long offset=0, high=0;

        user_data *u = (user_data*)malloc(MAX_USERS*sizeof(user_data));
        memset(u, 0, sizeof(user_data)*MAX_USERS);

        if (handle == NULL)
        {
                t_printmsg("Unable to open data file.");
                return;
        }

	
	for (;;)
        {
                fread(&u[i], sizeof(user_data), 1, handle);
		if (feof(handle)) break;
		if (u[i].win > high) high=u[i].win;
		i++;
	}
	printf("%d users.\n\r", i);
	k = 1;

	for (;high>=0; high--)
	for (j=0; j<i; j++)
		if (u[j].win==high)
		{
			total = u[i].win + u[i].lose + u[i].quit;
			if (k==1) printf(VT_BOLD);
			printf("%3d. %16s: won %4d (%3.1f%), lost %4d, quit %4d, total %4d\n\r", 
				k++, u[j].name, u[j].win, 
				(float)(u[j].win+1)/(float)(total+1)*100.0, 
				u[j].lose, u[j].quit, total);
			if (k==2) printf(VT_NORMAL);
		}

	free(u);
	fclose(handle);
}


void grid_render()
{
        int x,y;
        char line[81]; 
 
        printf("\x1B[0m\x1B[1;1H");

        for (y=0; y<GRIDYSIZE; y++)
        {
                for (x=0; x<GRIDXSIZE; x++)
                        line[x] = grid[x][y];

                line[80] = '\0';
                printf("%s\n\r", line);
        }
        fflush(stdout);
}

void draw_text(int x, int y, char *text)
{
        char buffer[256];
        sprintf(buffer, "\x1B[%d;%dH%s", y+1, x+1, text);
        printf(buffer);
        fflush(stdout);
}



void grid_draw_mark(int x, int y)
{
        int i;
        char text[8];

        printf("\x1B[0m");
        text[0] = grid[x][y];
        text[1] = '\0';

        draw_text(x, y, text);
        printf("\x1B[1D");
        fflush(stdout);
}

void grid_draw_inverse_mark(int x, int y)
{
        int i;
        char text[8];  

        printf("\x1B[1m");
        text[0] = grid[x][y];
        text[1] = '\0';

        draw_text(x, y, text);   
        printf("\x1B[1D");
        fflush(stdout);
}


int check_for_mark(int x, int y, char mark)
{
        if (x<0) return 0;
        if (y<0) return 0;
        if (x>GRIDXSIZE-1) return 0;
        if (y>GRIDYSIZE-1) return 0;
 
        if (grid[x][y]==mark) return 1;
        return 0;
}


char check_for_win()
{
        int x,y;
        char m;

        for (y=0; y<GRIDYSIZE; y++)
        for (x=0; x<GRIDXSIZE; x++)
        {
                m = grid[x][y];
                if (m != BLANKCHAR)
                {
                        if (check_for_mark(x+1, y, m) &&
                                        check_for_mark(x+2, y, m) &&
                                        check_for_mark(x+3, y, m) &&
                                        check_for_mark(x+4, y, m)) return m;

                        if (check_for_mark(x, y+1, m) &&
                                        check_for_mark(x, y+2, m) &&
                                        check_for_mark(x, y+3, m) &&
                                        check_for_mark(x, y+4, m)) return m;
 

                        if (check_for_mark(x+1, y+1, m) &&
                                        check_for_mark(x+2, y+2, m) &&
                                        check_for_mark(x+3, y+3, m) &&
                                        check_for_mark(x+4, y+4, m)) return m;

                        if (check_for_mark(x-1, y+1, m) &&
                                        check_for_mark(x-2, y+2, m) &&
                                        check_for_mark(x-3, y+3, m) &&
                                        check_for_mark(x-4, y+4, m)) return m;
                }
        }
                 
        return '\0';
}
                                        
                                        
void grid_init()
{
        memset(grid, BLANKCHAR, GRIDXSIZE*GRIDYSIZE);
}
                                        
int place_mark(char *mark)
{
        char key;
                                        
        grid_draw_inverse_mark(mark_x, mark_y);
                                        
        for (;;)
        {
                key = getch();
                grid_draw_mark(mark_x, mark_y);
                                        
                if (key == keyset[0]) mark_x--;
                if (key == keyset[1]) mark_x++;
                if (key == keyset[2]) mark_y--;
                if (key == keyset[3]) mark_y++;
                if (key == 'q') return 1;
                                        
                if (key == ' ')         
                {
                        if (grid[mark_x][mark_y]==BLANKCHAR)
                        {
                                grid[mark_x][mark_y] = mark[0];
                                grid_draw_mark(mark_x, mark_y);
                                return 0;
                        }
                }
                                        
                if (mark_x<0) mark_x=0;
                if (mark_y<0) mark_y=0; 

                if (mark_x>GRIDXSIZE-1) mark_x=GRIDXSIZE-1;
                if (mark_y>GRIDYSIZE-1) mark_y=GRIDYSIZE-1;
                
                grid_draw_inverse_mark(mark_x, mark_y);
        }
        return 0;
}
                
                
                                        
                

int ttt_play(int mode, int local, int remote)
{
	int i, ox, oy, state = 0;
	char buffer[80];

	t_mode = MODE_TTT;

/*	printf("Using keyset %s.\n\r", keyset); */

	cursor_set(1, 25);
        for (i=0; i<25; i++)
                printf("\n");
	fflush(stdout);

	cursor_set(56, 20);
	printf("IPCTTT by Sami Kystil\r\n");
	cursor_set(56, 21);
	printf("%c/%c/%c/%c/space/q\r\n", keyset[0], keyset[1], keyset[2], keyset[3]);
	
	cursor_set(1, 20);
	if (mode == M_SERVER)
	{
		printf("%sX: %s%s\n\r", VT_BOLD, me.name, VT_NORMAL); 
		printf("O: %s\n\r", r_user[remote].name);
	} else {
                printf("X: %s\n\r", r_user[remote].name);
                printf("%sO: %s%s\n\r", VT_BOLD, me.name, VT_NORMAL);
	}

	fflush(stdout);

	grid_init();
	grid_render();
	fflush(stdout);
	fflush(stdin);

	for (;;)
	{
		if (mode == M_SERVER)
		{
			cursor_set(1, 19);
			printf("Your turn      ");
			state = place_mark("X");
			if (!state)
			{
				sprintf(buffer, "%c%c", (char)mark_x+1, (char)mark_y+1);
/*				send_message(msgqueue_id, (struct mymsgbuf*)&qbuf, PLAYER_1, buffer); */
				msg_sendpriv(remote, M_TTTMOVE, buffer);
			} else {
				buffer[0] = 0xff;
				msg_sendpriv(remote, M_TTTMOVE, buffer);
                                cursor_set(1, 19);
                                printf("\x7User abort!         \n\r");
				update_userdata(me.name, 0, 0, 1);
                                break;
			}

			if (check_for_win() == 'X')
			{
                        	cursor_set(1, 19);
                        	printf("\x7You won!        \n\r");
				update_userdata(me.name, 1, 0, 0);
				break;
			}

                        cursor_set(1, 19);
                        printf("Opponent's turn");
                        fflush(stdout);
			last_msg_t = 0;
			while (last_msg_t != M_TTTMOVE) msg_handle();

                        ox = last_msg[0]-1;
                        oy = last_msg[1]-1;

                        if (ox==-2)
                        {
                                cursor_set(1, 19);
                                printf("\x7User abort!         \n\r");
                                break;
                        }


                        grid[ox][oy] = 'O';
                        grid_render();

                        if (check_for_win() == 'O')
                        {
                                cursor_set(1, 19);
                                printf("\x7You lost!         \n\r");
				update_userdata(me.name, 0, 1, 0);
                                break;
                        }
		} else {
			cursor_set(1, 19);
			printf("Opponent's turn");
			fflush(stdout);

			last_msg_t = 0;
			while (last_msg_t != M_TTTMOVE) msg_handle();

                        ox = last_msg[0]-1;
                        oy = last_msg[1]-1;

/*
                	read_message(msgqueue_id, &qbuf, PLAYER_1);
			ox = qbuf.mtext[0]-1;
			oy = qbuf.mtext[1]-1;
*/
			if (ox==-2)
			{
                                cursor_set(1, 19);
                                printf("\x7User abort!         \n\r");
                                break;
			}

			grid[ox][oy] = 'X';
			grid_render();
                        if (check_for_win() == 'X')
                        {
                                cursor_set(1, 19);
                                printf("\x7You lost!         \n\r");
				update_userdata(me.name, 0, 1, 0);
                                break;
                        }

			
                        cursor_set(1, 19);
                        printf("Your turn      ");
                        state = place_mark("O");
                        if (!state)
                        {
                                sprintf(buffer, "%c%c", (char)mark_x+1, (char)mark_y+1);
/*                                send_message(msgqueue_id, (struct mymsgbuf*)&qbuf, PLAYER_2, buffer);*/
				msg_sendpriv(remote, M_TTTMOVE, buffer);
                        } else {
                                buffer[0] = 0xff;
				msg_sendpriv(remote, M_TTTMOVE, buffer);
/*
                                send_message(msgqueue_id, (struct mymsgbuf*)&qbuf, PLAYER_2, buffer);		
*/
                                cursor_set(1, 19);
                                printf("\x7User abort!         \n\r");
				update_userdata(me.name, 0, 0, 1);
                                break;
			}
			if (check_for_win() == 'O')
                        {
                                cursor_set(1, 19);
                                printf("\x7You won!        \n\r");
				update_userdata(me.name, 1, 0, 0);
                                break;
                        }

		}
		if (state) break;
	}
	t_mode = MODE_NORMAL;
	return(0);
}


